#!/usr/bin/env python
# pylint:disable=W0404,W0612,C0111

"""Integration tests the package namespace."""

import pytest


def test_top():

    import yorm

    assert yorm.base
    assert yorm.converters.Integer
    assert yorm.converters.extended.Markdown


def test_from_top():

    # Constants
    from yorm import UUID  # filename placeholder

    # Classes
    from yorm import Mappable  # base class for mapped objects
    from yorm import Converter, Convertible  # base class for converters

    # Decorators
    from yorm import sync  # enables mapping on a class's instance objects
    from yorm import sync_instances  # alias for the class decorator
    from yorm import attr  # alternate API to identify mapped attributes

    # Functions
    from yorm import sync  # enables mapping on an instance object
    from yorm import sync_object  # alias for the mapping function
    from yorm import update  # fetch (if necessary) and store a mapped object
    from yorm import update_object  # fetch (optional force) a mapped object
    from yorm import update_file  # store a mapped object


def test_from_nested():

    # Classes
    from yorm.converters import Integer
    from yorm.converters.standard import String
    from yorm.converters.extended import Markdown
    from yorm.converters.containers import List


if __name__ == '__main__':
    pytest.main()



"""Integration tests for the `yorm` package."""

import time
import logging


def strip(text, tabs=None, end='\n'):
    """Strip leading whitespace indentation on multiline string literals."""
    lines = text.strip().splitlines()
    for index in range(len(lines)):
        if not tabs:
            tabs = lines[index].count(' ' * 4)
        lines[index] = lines[index].replace(' ' * tabs * 4, '')
    return '\n'.join(lines) + end


def refresh_file_modification_times(seconds=1.1):
    """Sleep to allow file modification times to refresh."""
    logging.info("delaying for %s second%s...", seconds,
                 "" if seconds == 1 else "s")
    time.sleep(seconds)



#!/usr/bin/env python
# pylint:disable=W0201,W0613,R0201,W0212,C0111

"""Unit tests for the `base.convertible` module."""

import pytest

from yorm.base.convertible import Convertible


class TestConvertible:

    """Unit tests for the `Convertible` class."""

    class MyConvertible(Convertible):

        def __init__(self, number):
            self.value = number

        @classmethod
        def create_default(cls):
            return 1

        @classmethod
        def to_data(cls, value):
            return str(value.value)

        def update_value(self, data):
            self.value += int(data)

    def test_convertible_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            Convertible()  # pylint: disable=E0110

    def test_convertible_instance_methods_can_be_called(self):
        convertible = self.MyConvertible(42)
        assert 42 == convertible.value
        convertible.update_value(10)
        assert 52 == convertible.value
        assert "52" == convertible.format_data()


if __name__ == '__main__':
    pytest.main()



#!/usr/bin/env python
# pylint:disable=R0201,C0111

"""Integration tests for the package."""

import pytest

from yorm import sync
from yorm.converters import Object, String, Integer, Float, Boolean
from yorm.converters import Markdown

from . import strip, refresh_file_modification_times
from .samples import *  # pylint: disable=W0401,W0614


class TestStandard:

    """Integration tests for standard attribute types."""

    class EmptyDictionary(Dictionary):

        pass

    def test_decorator(self, tmpdir):
        """Verify standard attribute types dump/load correctly (decorator)."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        assert "path/to/default/sample.yml" == sample.yorm_mapper.path

        # check defaults
        assert {} == sample.object
        assert [] == sample.array
        assert "" == sample.string
        assert 0 == sample.number_int
        assert 0.0 == sample.number_real
        assert True is sample.true
        assert False is sample.false
        assert None is sample.null

        # change object values
        sample.object = {'key2': 'value'}
        sample.array = [0, 1, 2]
        sample.string = "Hello, world!"
        sample.number_int = 42
        sample.number_real = 4.2
        sample.true = False
        sample.false = True

        # check file values
        assert strip("""
        array:
        - 0
        - 1
        - 2
        'false': true
        number_int: 42
        number_real: 4.2
        object:
          key2: value
        string: Hello, world!
        'true': false
        """) == sample.yorm_mapper.text

        # change file values
        refresh_file_modification_times()
        sample.yorm_mapper.text = strip("""
        array: [4, 5, 6]
        'false': null
        number_int: 42
        number_real: '4.2'
        object: {'status': false}
        string: "abc"
        'true': null
        """)

        # check object values
        assert {'key2': "",
                'status': False} == sample.object
        assert [4, 5, 6] == sample.array
        assert "abc" == sample.string
        assert 42 == sample.number_int
        assert 4.2 == sample.number_real
        assert False is sample.true
        assert False is sample.false

    def test_function(self, tmpdir):
        """Verify standard attribute types dump/load correctly (function)."""
        tmpdir.chdir()
        _sample = SampleStandard()
        attrs = {'object': self.EmptyDictionary,
                 'array': IntegerList,
                 'string': String,
                 'number_int': Integer,
                 'number_real': Float,
                 'true': Boolean,
                 'false': Boolean}
        sample = sync(_sample, "path/to/directory/sample.yml", attrs)
        assert "path/to/directory/sample.yml" == sample.yorm_mapper.path

        # check defaults
        assert {} == sample.object
        assert [] == sample.array
        assert "" == sample.string
        assert 0 == sample.number_int
        assert 0.0 == sample.number_real
        assert True is sample.true
        assert False is sample.false
        assert None is sample.null

        # change object values
        sample.object = {'key': 'value'}
        sample.array = [1, 2, 3]
        sample.string = "Hello, world!"
        sample.number_int = 42
        sample.number_real = 4.2
        sample.true = None
        sample.false = 1

        # check file values
        assert strip("""
        array:
        - 1
        - 2
        - 3
        'false': true
        number_int: 42
        number_real: 4.2
        object:
          key: value
        string: Hello, world!
        'true': false
        """) == sample.yorm_mapper.text

    def test_auto_off(self, tmpdir):
        """Verify file updates are disabled with auto off."""
        tmpdir.chdir()
        sample = SampleDecoratedAutoOff()

        # check for default values
        assert strip("""
        string: ''
        """) == sample.yorm_mapper.text

        # store value
        sample.string = "hello"

        # check for unchanged file values
        assert strip("""
        string: ''
        """) == sample.yorm_mapper.text

        # enable auto and store value
        sample.yorm_mapper.auto = True
        sample.string = "world"

        # check for changed file values
        assert strip("""
        string: world
        """) == sample.yorm_mapper.text


class TestContainers:

    """Integration tests for attribute containers."""

    def test_nesting(self, tmpdir):
        """Verify standard attribute types can be nested."""
        tmpdir.chdir()
        _sample = SampleNested()
        attrs = {'count': Integer,
                 'results': StatusDictionaryList}
        sample = sync(_sample, "sample.yml", attrs)

        # check defaults
        assert 0 == sample.count
        assert [] == sample.results

        # change object values
        sample.count = 5
        sample.results = [{'status': False, 'label': "abc"},
                          {'status': None, 'label': None},
                          {'label': "def"},
                          {'status': True},
                          {}]

        # check file values
        assert strip("""
        count: 5
        results:
        - label: abc
          status: false
        - label: ''
          status: false
        - label: def
          status: false
        - label: ''
          status: true
        - label: ''
          status: false
        """) == sample.yorm_mapper.text

        # change file values
        refresh_file_modification_times()
        sample.yorm_mapper.text = strip("""
        count: 3
        other: 4.2
        results:
        - label: abc
        - label: null
          status: false
        - status: true
        """)

        # check object values
        assert 3 == sample.count
        assert 4.2 == sample.other
        assert [{'label': 'abc', 'status': False},
                {'label': '', 'status': False},
                {'label': '', 'status': True}] == sample.results

    def test_objects(self, tmpdir):
        """Verify containers are treated as objects when added."""
        tmpdir.chdir()
        sample = SampleEmptyDecorated()

        # change file values
        refresh_file_modification_times()
        sample.yorm_mapper.text = strip("""
        object: {'key': 'value'}
        array: [1, '2', '3.0']
        """)

        # (a mapped attribute must be read first to trigger retrieving)
        sample.yorm_mapper.fetch()

        # check object values
        assert {'key': 'value'} == sample.object
        assert [1, '2', '3.0'] == sample.array

        # check object types
        assert Object == sample.yorm_mapper.attrs['object']
        assert Object == sample.yorm_mapper.attrs['array']

        # change object values
        sample.object = None  # pylint: disable=W0201
        sample.array = "abc"  # pylint: disable=W0201

        # check file values
        assert strip("""
        array: abc
        object: null
        """) == sample.yorm_mapper.text


class TestExtended:

    """Integration tests for extended attribute types."""

    def test_function(self, tmpdir):
        """Verify extended attribute types dump/load correctly."""
        tmpdir.chdir()
        _sample = SampleExtended()
        attrs = {'text': Markdown}
        sample = sync(_sample, "path/to/directory/sample.yml", attrs)

        # check defaults
        assert "" == sample.text

        # change object values
        refresh_file_modification_times()
        sample.text = strip("""
        This is the first sentence. This is the second sentence.
        This is the third sentence.
        """)

        # check file values
        assert strip("""
        text: |
          This is the first sentence.
          This is the second sentence.
          This is the third sentence.
        """) == sample.yorm_mapper.text

        # change file values
        refresh_file_modification_times()
        sample.yorm_mapper.text = strip("""
        text: |
          This is a
          sentence.
        """)

        # check object values
        assert "This is a sentence." == sample.text


class TestCustom:

    """Integration tests for custom attribute types."""

    def test_decorator(self, tmpdir):
        """Verify custom attribute types dump/load correctly."""
        tmpdir.chdir()
        sample = SampleCustomDecorated('sample')

        # check defaults
        assert '1.0' == sample.level

        # change values
        sample.level = '1.2.3'

        # check file values
        assert strip("""
        level: 1.2.3
        """) == sample.yorm_mapper.text

        # change file values
        refresh_file_modification_times()
        sample.yorm_mapper.text = strip("""
        level: 1
        """)

        # check object values
        assert '1' == sample.level


if __name__ == '__main__':
    pytest.main()



"""Base classes for conversion."""

import abc

from .. import common
from . import MESSAGE
from .converter import Converter


log = common.logger(__name__)


class Convertible(Converter):  # pragma: no cover (abstract)

    """Base class for mutable attributes."""

    @classmethod
    def to_value(cls, data):
        value = cls.create_default()
        value.update_value(data)
        return value

    @abc.abstractmethod
    def update_value(self, data):  # pylint: disable=E0213,
        """Update the attribute's value from loaded data."""
        raise NotImplementedError(MESSAGE)

    def format_data(self):
        """Format the attribute to data optimized for dumping."""
        return self.to_data(self)



"""Base classes for mapping."""

import abc
import functools

from .. import common

log = common.logger(__name__)


MAPPER = 'yorm_mapper'


def get_mapper(obj):
    """Get `Mapper` instance attached to an object."""
    if not isinstance(obj, Mappable):
        msg = "object {!r} must be a `Mappable` instance".format(obj)
        raise TypeError(msg)
    try:
        mapper = getattr(obj, MAPPER)
    except AttributeError:
        if isinstance(obj, (dict, list)):
            return None
        else:
            msg = "mapped {!r} missing {!r} attribute".format(obj, MAPPER)
            raise AttributeError(msg) from None
    else:
        return mapper


def fetch_before(method):
    """Decorator for methods that should fetch before call."""
    @functools.wraps(method)
    def fetch_before(self, *args, **kwargs):  # pylint: disable=W0621
        """Decorated method."""
        mapper = get_mapper(self)
        if mapper:
            log.debug("fetch before call: %s", method.__name__)
            mapper.fetch()
        return method(self, *args, **kwargs)
    return fetch_before


def store_after(method):
    """Decorator for methods that should store after call."""
    @functools.wraps(method)
    def store_after(self, *args, **kwargs):  # pylint: disable=W0621
        """Decorated method."""
        result = method(self, *args, **kwargs)
        mapper = get_mapper(self)
        if mapper:
            log.debug("store after call: %s", method.__name__)
            mapper.store()
        return result
    return store_after


class Mappable(metaclass=abc.ABCMeta):  # pylint: disable=R0201

    """Base class for objects with attributes mapped to file."""

    def __getattribute__(self, name):
        """Trigger object update when reading attributes."""
        if name in ('__dict__', '__class__', MAPPER):
            # avoid infinite recursion (attribute requested in this function)
            return object.__getattribute__(self, name)

        mapper = get_mapper(self)

        # Get the attribute's current value
        try:
            value = object.__getattribute__(self, name)
        except AttributeError as exc:
            missing = True
            if not mapper:
                raise exc from None
        else:
            missing = False

        # Fetch a new value from disk if the attribute is mapped or missing
        if mapper and (missing or name in mapper.attrs):
            mapper.fetch()
            value = object.__getattribute__(self, name)

        return value

    def __setattr__(self, name, value):
        """Trigger file update when setting attributes."""
        super().__setattr__(name, value)
        if name in ('__dict__', MAPPER):
            return

        mapper = get_mapper(self)
        if mapper and name in mapper.attrs:
            mapper.store()

    @fetch_before
    def __iter__(self):
        """Trigger object update when iterating."""
        return super().__iter__()

    @fetch_before
    def __getitem__(self, key):
        """Trigger object update when reading an index."""
        return super().__getitem__(key)

    @store_after
    def __setitem__(self, key, value):
        """Trigger file update when setting an index."""
        super().__setitem__(key, value)

    @store_after
    def __delitem__(self, key):
        """Trigger file update when deleting an index."""
        super().__delitem__(key)

    @store_after
    def append(self, value):
        """Trigger file update when appending items."""
        super().append(value)



#!/usr/bin/env python
# pylint:disable=R0201,C0111

"""Integration tests for file IO."""

import logging

import pytest

from . import refresh_file_modification_times
from .samples import *  # pylint: disable=W0401,W0614


class TestInit:

    """Integration tests for initializing mapped classes."""

    def test_fetch_from_existing(self, tmpdir):
        """Verify attributes are updated from an existing file."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample2 = SampleStandardDecorated('sample')
        assert sample2.yorm_mapper.path == sample.yorm_mapper.path

        refresh_file_modification_times()

        logging.info("changing values in object 1...")
        sample.object = {'key2': 'value'}
        sample.array = [0, 1, 2]
        sample.string = "Hello, world!"
        sample.number_int = 42
        sample.number_real = 4.2
        sample.true = True
        sample.false = False

        logging.info("reading changed values in object 2...")
        assert 'value' == sample2.object.get('key2')
        assert [0, 1, 2] == sample2.array
        assert "Hello, world!" == sample2.string
        assert 42 == sample2.number_int
        assert 4.2 == sample2.number_real
        assert True is sample2.true
        assert False is sample2.false


class TestDelete:

    """Integration tests for deleting files."""

    def test_read(self, tmpdir):
        """Verify a deleted file cannot be read from."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample.yorm_mapper.delete()

        with pytest.raises(FileNotFoundError):
            print(sample.string)

        with pytest.raises(FileNotFoundError):
            sample.string = "def456"

    def test_write(self, tmpdir):
        """Verify a deleted file cannot be written to."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample.yorm_mapper.delete()

        with pytest.raises(FileNotFoundError):
            sample.string = "def456"

    def test_multiple(self, tmpdir):
        """Verify a deleted file can be deleted again."""
        tmpdir.chdir()
        sample = SampleStandardDecorated('sample')
        sample.yorm_mapper.delete()
        sample.yorm_mapper.delete()


if __name__ == '__main__':
    pytest.main()



"""Package for YORM."""

import sys

__project__ = 'YORM'
__version__ = '0.4.1rc2'

VERSION = __project__ + '-' + __version__

PYTHON_VERSION = 3, 3

if not sys.version_info >= PYTHON_VERSION:  # pragma: no cover (manual test)
    exit("Python {}.{}+ is required.".format(*PYTHON_VERSION))

try:
    from . import base, converters
    from .utilities import UUID
    from .utilities import sync, sync_object, sync_instances, attr
    from .utilities import update, update_object, update_file
    from .base.mappable import Mappable
    from .base.convertible import Converter, Convertible
except ImportError:  # pragma: no cover (manual test)
    pass



#!/usr/bin/env python
# pylint:disable=R0201

"""Unit tests for the `utilities` module."""

import pytest
from unittest.mock import patch, Mock

from yorm import common
from yorm import utilities
from yorm.base.converter import Converter
from yorm.base.mappable import Mappable


class MockConverter(Converter):

    """Sample converter class."""

    @classmethod
    def create_default(cls):
        return None

    @classmethod
    def to_value(cls, _):
        return None

    @classmethod
    def to_data(cls, _):
        return None


class MockConverter0(MockConverter):

    """Sample converter class."""


class MockConverter1(MockConverter):

    """Sample converter class."""


class MockConverter2(MockConverter):

    """Sample converter class."""


class MockConverter3(MockConverter):

    """Sample converter class."""


class MockConverter4(MockConverter):

    """Sample converter class."""


class MockMappable(Mappable):

    """Sample mappable class."""

    yorm_mapper = Mock()
    yorm_mapper.attrs = {}


@patch('yorm.common.write_text', Mock())
@patch('yorm.common.stamp', Mock())
@patch('yorm.common.read_text', Mock(return_value=""))
class TestSyncObject:

    """Unit tests for the `sync_object` function."""

    class Sample:

        """Sample class."""

    def test_no_attrs(self):
        """Verify mapping can be enabled with no attributes."""
        sample = utilities.sync(self.Sample(), "sample.yml")
        assert "sample.yml" == sample.yorm_mapper.path
        assert {} == sample.yorm_mapper.attrs

    def test_with_attrs(self):
        """Verify mapping can be enabled with with attributes."""
        attrs = {'var1': MockConverter}
        sample = utilities.sync(self.Sample(), "sample.yml", attrs)
        assert "sample.yml" == sample.yorm_mapper.path
        assert {'var1': MockConverter} == sample.yorm_mapper.attrs

    def test_multiple(self):
        """Verify mapping cannot be enabled twice."""
        sample = utilities.sync(self.Sample(), "sample.yml")
        with pytest.raises(common.UseageError):
            utilities.sync(sample, "sample.yml")

    @patch('os.path.isfile', Mock(return_value=True))
    def test_init_existing(self):
        """Verify an existing file is read."""
        with patch('yorm.common.read_text', Mock(return_value="abc: 123")):
            sample = utilities.sync(self.Sample(), "sample.yml")
        assert 123 == sample.abc


@patch('yorm.common.create_dirname', Mock())
@patch('yorm.common.write_text', Mock())
@patch('yorm.common.stamp', Mock())
@patch('yorm.common.read_text', Mock(return_value=""))
class TestSyncInstances:

    """Unit tests for the `sync_instances` decorator."""

    @utilities.sync("sample.yml")
    class SampleDecorated:

        """Sample decorated class using a single path."""

        def __repr__(self):
            return "<decorated {}>".format(id(self))

    @utilities.sync("{UUID}.yml")
    class SampleDecoratedIdentifiers:

        """Sample decorated class using UUIDs for paths."""

        def __repr__(self):
            return "<decorated w/ UUID {}>".format(id(self))

    @utilities.sync("path/to/{n}.yml", {'n': 'name'})
    class SampleDecoratedAttributes:

        """Sample decorated class using an attribute value for paths."""

        def __init__(self, name):
            self.name = name

        def __repr__(self):
            return "<decorated w/ specified attributes {}>".format(id(self))

    @utilities.sync("path/to/{self.name}.yml")
    class SampleDecoratedAttributesAutomatic:

        """Sample decorated class using an attribute value for paths."""

        def __init__(self, name):
            self.name = name

        def __repr__(self):
            return "<decorated w/ automatic attributes {}>".format(id(self))

    @utilities.sync("{self.a}/{self.b}/{c}.yml", {'self.b': 'b', 'c': 'c'})
    class SampleDecoratedAttributesCombination:

        """Sample decorated class using an attribute value for paths."""

        def __init__(self, a, b, c):
            self.a = a
            self.b = b
            self.c = c

        def __repr__(self):
            return "<decorated w/ attributes {}>".format(id(self))

    @utilities.sync("sample.yml", attrs={'var1': MockConverter})
    class SampleDecoratedWithAttributes:

        """Sample decorated class using a single path."""

    @utilities.sync("sample.yml", attrs={'var1': MockConverter}, auto=False)
    class SampleDecoratedWithAttributesAutoOff:

        """Sample decorated class using a single path."""

    def test_no_attrs(self):
        """Verify mapping can be enabled with no attributes."""
        sample = self.SampleDecorated()
        assert "sample.yml" == sample.yorm_mapper.path
        assert {} == sample.yorm_mapper.attrs

    def test_with_attrs(self):
        """Verify mapping can be enabled with with attributes."""
        sample = self.SampleDecoratedWithAttributes()
        assert "sample.yml" == sample.yorm_mapper.path
        assert ['var1'] == list(sample.yorm_mapper.attrs.keys())

    @patch('os.path.isfile', Mock(return_value=True))
    def test_init_existing(self):
        """Verify an existing file is read."""
        with patch('yorm.common.read_text', Mock(return_value="abc: 123")):
            sample = self.SampleDecorated()
        assert 123 == sample.abc

    @patch('uuid.uuid4', Mock(return_value=Mock(hex='abc123')))
    def test_filename_uuid(self):
        """Verify UUIDs can be used for filename."""
        sample = self.SampleDecoratedIdentifiers()
        assert "abc123.yml" == sample.yorm_mapper.path
        assert {} == sample.yorm_mapper.attrs

    def test_filename_attributes(self):
        """Verify attributes can be used to determine filename."""
        sample1 = self.SampleDecoratedAttributes('one')
        sample2 = self.SampleDecoratedAttributes('two')
        assert "path/to/one.yml" == sample1.yorm_mapper.path
        assert "path/to/two.yml" == sample2.yorm_mapper.path

    def test_filename_attributes_automatic(self):
        """Verify attributes can be used to determine filename (auto)."""
        sample1 = self.SampleDecoratedAttributesAutomatic('one')
        sample2 = self.SampleDecoratedAttributesAutomatic('two')
        assert "path/to/one.yml" == sample1.yorm_mapper.path
        assert "path/to/two.yml" == sample2.yorm_mapper.path

    def test_filename_attributes_combination(self):
        """Verify attributes can be used to determine filename (combo)."""
        sample1 = self.SampleDecoratedAttributesCombination('A', 'B', 'C')
        sample2 = self.SampleDecoratedAttributesCombination(1, 2, 3)
        assert "A/B/C.yml" == sample1.yorm_mapper.path
        assert "1/2/3.yml" == sample2.yorm_mapper.path


@patch('yorm.common.write_text', Mock())
@patch('yorm.common.stamp', Mock())
@patch('yorm.common.read_text', Mock(return_value=""))
class TestAttr:

    """Unit tests for the `attr` decorator."""

    @utilities.attr(var1=MockConverter1, var2=MockConverter2)
    @utilities.sync("sample.yml")
    class SampleDecoratedSingle:

        """Sample decorated class using one `attr` decorator."""

    @utilities.attr()
    @utilities.attr(var1=MockConverter1)
    @utilities.attr(var2=MockConverter2, var3=MockConverter3)
    @utilities.sync("sample.yml")
    class SampleDecoratedMultiple:

        """Sample decorated class using many `attr` decorators."""

    @utilities.attr()
    @utilities.attr(var1=MockConverter1)
    @utilities.attr(var2=MockConverter2, var3=MockConverter3)
    @utilities.sync("sample.yml", attrs={'var0': MockConverter0})
    class SampleDecoratedCombo:

        """Sample decorated class using `attr` and providing a mapping."""

    @utilities.sync("sample.yml", attrs={'var0': MockConverter0})
    @utilities.attr(var1=MockConverter1)
    class SampleDecoratedBackwards:

        """Sample decorated class using one `attr` decorator."""

    def test_single(self):
        """Verify `attr` can be applied once."""
        sample = self.SampleDecoratedSingle()
        expected = {'var1': MockConverter1,
                    'var2': MockConverter2}
        assert expected == sample.yorm_mapper.attrs

    def test_multiple(self):
        """Verify `attr` can be applied many times."""
        sample = self.SampleDecoratedMultiple()
        expected = {'var1': MockConverter1,
                    'var2': MockConverter2,
                    'var3': MockConverter3}
        assert expected == sample.yorm_mapper.attrs

    def test_combo(self):
        """Verify `attr` can be applied an existing mapping."""
        sample = self.SampleDecoratedCombo()
        expected = {'var0': MockConverter0,
                    'var1': MockConverter1,
                    'var2': MockConverter2,
                    'var3': MockConverter3}
        assert expected == sample.yorm_mapper.attrs

    def test_backwards(self):
        """Verify `attr` can be applied before `sync`."""
        sample = self.SampleDecoratedBackwards()
        expected = {'var0': MockConverter0,
                    'var1': MockConverter1}
        assert expected == sample.yorm_mapper.attrs


class TestUpdate:

    """Unit tests for the `update` function."""

    def test_update(self):
        """Verify the object and file are updated."""
        instance = MockMappable()
        instance.yorm_mapper.reset_mock()

        utilities.update(instance)

        assert instance.yorm_mapper.fetch.called
        assert instance.yorm_mapper.store.called

    def test_update_object_only(self):
        """Verify only the object is updated."""
        instance = MockMappable()
        instance.yorm_mapper.reset_mock()

        utilities.update(instance, store=False)

        assert instance.yorm_mapper.fetch.called
        assert not instance.yorm_mapper.store.called

    def test_update_file_only(self):
        """Verify only the file is updated."""
        instance = MockMappable()
        instance.yorm_mapper.reset_mock()

        utilities.update(instance, fetch=False)

        assert not instance.yorm_mapper.fetch.called
        assert instance.yorm_mapper.store.called

    def test_update_wrong_base(self):
        """Verify an exception is raised with the wrong base."""
        instance = Mock()

        with pytest.raises(common.UseageError):
            utilities.update(instance)


class TestUpdateObject:

    """Unit tests for the `update_object` function."""

    def test_update(self):
        """Verify only the object is updated."""
        instance = MockMappable()
        instance.yorm_mapper.reset_mock()

        utilities.update_object(instance)

        assert instance.yorm_mapper.fetch.called
        assert not instance.yorm_mapper.store.called

    def test_update_wrong_base(self):
        """Verify an exception is raised with the wrong base."""
        instance = Mock()

        with pytest.raises(common.UseageError):
            utilities.update_object(instance)


class TestUpdateFile:

    """Unit tests for the `update_file` function."""

    def test_update(self):
        """Verify only the file is updated."""
        instance = MockMappable()
        instance.yorm_mapper.reset_mock()

        utilities.update_file(instance)

        assert not instance.yorm_mapper.fetch.called
        assert instance.yorm_mapper.store.called

    def test_update_wrong_base(self):
        """Verify an exception is raised with the wrong base."""
        instance = Mock()

        with pytest.raises(common.UseageError):
            utilities.update_file(instance)


if __name__ == '__main__':
    pytest.main()



#!/usr/bin/env python
# pylint:disable=W0201,W0613,R0201,W0212,C0111

"""Unit tests for the `base.converter` module."""

import pytest

from yorm.base.converter import Converter


class TestConverter:

    """Unit tests for the `Converter` class."""

    def test_converter_class_cannot_be_instantiated(self):
        with pytest.raises(TypeError):
            Converter()  # pylint: disable=E0110

    def test_converter_class_methods_cannot_be_called(self):
        with pytest.raises(NotImplementedError):
            Converter.create_default()
        with pytest.raises(NotImplementedError):
            Converter.to_value(None)
        with pytest.raises(NotImplementedError):
            Converter.to_data(None)


if __name__ == '__main__':
    pytest.main()



"""Base classes for conversion."""

import abc

from .. import common
from . import MESSAGE


log = common.logger(__name__)


class Converter(metaclass=abc.ABCMeta):

    """Base class for immutable attribute converters."""

    @abc.abstractclassmethod
    def create_default(cls):
        """Create a default value for an attribute."""
        raise NotImplementedError(MESSAGE)

    @abc.abstractclassmethod
    def to_value(cls, data):
        """Convert loaded data to an attribute's value."""
        raise NotImplementedError(MESSAGE)

    @abc.abstractclassmethod
    def to_data(cls, value):
        """Convert an attribute to data optimized for dumping."""
        raise NotImplementedError(MESSAGE)



"""Core YAML mapping functionality."""

import os
import abc
import functools

import yaml

from . import common
from . import settings
from .base.mappable import MAPPER, Mappable
from .base.convertible import Convertible

log = common.logger(__name__)


def file_required(method):
    """Decorator for methods that require the file to exist."""
    @functools.wraps(method)
    def file_required(self, *args, **kwargs):  # pylint: disable=W0621
        """Decorated method."""
        if not self.path:
            return None
        if not self.exists:
            msg = "cannot access deleted: {}".format(self.path)
            raise common.FileError(msg)
        return method(self, *args, **kwargs)
    return file_required


def prefix(obj):
    """Prefix a string with a fake designator if enabled."""
    fake = "(fake) " if settings.fake else ""
    name = obj if isinstance(obj, str) else "'{}'".format(obj)
    return fake + name


class BaseHelper(metaclass=abc.ABCMeta):

    """Utility class to map attributes to text files.

    To start mapping attributes to a file:

        create -> [empty] -> FILE

    When getting an attribute:

        FILE -> read -> [text] -> load -> [dict] -> fetch -> ATTRIBUTES

    When setting an attribute:

        ATTRIBUTES -> store -> [dict] -> dump -> [text] -> write -> FILE

    After the mapped file is no longer needed:

        delete -> [null] -> FILE

    """

    def __init__(self, path, auto=True):
        self.path = path
        self.auto = auto
        self.exists = self.path and os.path.isfile(self.path)
        self._activity = False
        self._timestamp = 0
        self._fake = ""

    def __str__(self):
        return str(self.path)

    @property
    def text(self):
        """Get file contents."""
        log.info("getting contents of %s...", prefix(self))
        if settings.fake:
            text = self._fake
        else:
            text = self._read()
        log.trace("text read: %r", text)
        return text

    @text.setter
    def text(self, text):
        """Set file contents."""
        log.info("setting contents of %s...", prefix(self))
        if settings.fake:
            self._fake = text
        else:
            self._write(text)
        log.trace("text wrote: %r", text)
        self.modified = True

    def create(self, obj):
        """Create a new file for the object."""
        log.info("creating %s for %r...", prefix(self), obj)
        if self.exists:
            log.warning("already created: %s", self)
            return
        if not settings.fake:
            common.create_dirname(self.path)
            common.touch(self.path)
        self.exists = True

    @file_required
    def fetch(self, obj, attrs, force=False):
        """Update the object's mapped attributes from its file."""
        if self._activity:
            return
        if not self.modified and not force:
            return
        self._activity = True
        _force = "force-" if force else ""
        log.info("%sfetching %r from %s...", _force, obj, prefix(self))

        # Parse data from file
        text = self._read()
        data = self._load(text, self.path)
        log.trace("loaded: {}".format(data))

        # Update all attributes
        for name, data in data.items():

            # Find a matching converter
            try:
                converter = attrs[name]
            except KeyError:
                # TODO: determine if runtime import is the best way to avoid cyclic import
                from .converters import match
                converter = match(name, data)
                attrs[name] = converter

            # Convert the loaded attribute
            attr = getattr(obj, name, None)
            if all((isinstance(attr, converter),
                    issubclass(converter, Convertible))):
                attr.update_value(data)
            else:
                attr = converter.to_value(data)
                setattr(obj, name, attr)
            self._remap(attr)
            log.trace("value fetched: '%s' = %r", name, attr)

        # Set meta attributes
        self.modified = False
        self._activity = False

    def _remap(self, obj):
        """Restore mapping on nested attributes."""
        if isinstance(obj, Mappable):
            mapper = Mapper(obj, None, common.ATTRS[obj.__class__], root=self)
            setattr(obj, MAPPER, mapper)

            if isinstance(obj, dict):
                for obj2 in obj.values():
                    self._remap(obj2)
            else:
                assert isinstance(obj, list)
                for obj2 in obj:
                    self._remap(obj2)

    @file_required
    def _read(self):
        """Read text from the object's file.

        :param path: path to a text file

        :return: contexts of text file

        """
        if settings.fake:
            return self._fake
        else:
            return common.read_text(self.path)

    @abc.abstractstaticmethod
    def _load(text, path):  # pragma: no cover (abstract method)
        """Parsed data from text.

        :param text: text read from a file
        :param path: path to the file (for displaying errors)

        :return: dictionary of parsed data

        """
        raise NotImplementedError

    @file_required
    def store(self, obj, attrs, force=False):
        """Format and save the object's mapped attributes to its file."""
        if self._activity:
            return
        if not self.auto and not force:
            log.debug("automatic storage is off ")
            return
        self._activity = True
        _force = "force-" if force else ""
        log.info("%sstoring %r to %s...", _force, obj, prefix(self))

        # Format the data items
        data = {}
        for name, converter in attrs.items():
            try:
                value = getattr(obj, name)
            except AttributeError:
                log.warn("added missing attribute '%s'", name)
                value = None

            data2 = converter.to_data(value)

            log.trace("data to store: '%s' = %r", name, data2)
            data[name] = data2

        # Dump data to file
        text = self._dump(data)
        self._write(text)

        # Set meta attributes
        self.modified = True
        self._activity = False

    @abc.abstractstaticmethod
    def _dump(data):  # pragma: no cover (abstract method)
        """Dump data to text.

        :param data: dictionary of data

        :return: text to write to a file

        """
        raise NotImplementedError

    @file_required
    def _write(self, text):
        """Write text to the object's file.

        :param text: text to write to a file
        :param path: path to the file

        """
        if settings.fake:
            self._fake = text
        else:
            common.write_text(text, self.path)

    @property
    def modified(self):
        """Determine if the file has been modified."""
        if settings.fake:
            changes = self._timestamp is not None
            return changes
        elif not self.exists:
            return True
        else:
            was = self._timestamp
            now = common.stamp(self.path)
            return was != now

    @modified.setter
    def modified(self, changes):
        """Mark the file as modified if there are changes."""
        if changes:
            log.debug("marked %s as modified", prefix(self))
            self._timestamp = 0
        else:
            if settings.fake:
                self._timestamp = None
            else:
                self._timestamp = common.stamp(self.path)
            log.debug("marked %s as unmodified", prefix(self))

    def delete(self):
        """Delete the object's file from the file system."""
        if self.exists:
            log.info("deleting %s...", prefix(self))
            if not settings.fake:
                common.delete(self.path)
            self.exists = False
        else:
            log.warning("already deleted: %s", self)


class Helper(BaseHelper):

    """Utility class to map attributes to YAML files."""

    @staticmethod
    def _load(text, path):
        """Load YAML data from text.

        :param text: text read from a file
        :param path: path to the file (for displaying errors)

        :return: dictionary of YAML data

        """
        return common.load_yaml(text, path)

    @staticmethod
    def _dump(data):
        """Dump YAML data to text.

        :param data: dictionary of YAML data

        :return: text to write to a file

        """
        return yaml.dump(data, default_flow_style=False, allow_unicode=True)


class Mapper(Helper):

    """Maps an object's attribute to YAML files."""

    def __init__(self, obj, path, attrs, auto=True, root=None):  # pylint: disable=R0913
        super().__init__(path, auto=auto)
        self.obj = obj
        self.attrs = attrs
        self.root = root

    def create(self):  # pylint: disable=W0221
        super().create(self.obj)

    def fetch(self, force=False):  # pylint: disable=W0221
        if self.root and self.root.auto and not self._activity:
            self._activity = True
            self.root.fetch(force=force)
            self._activity = False
        super().fetch(self.obj, self.attrs, force=force)

    def store(self, force=False):  # pylint: disable=W0221
        if self.root and self.root.auto and not self._activity:
            self._activity = True
            self.root.store(force=force)
            self._activity = False
        super().store(self.obj, self.attrs, force=force)



"""Converters for attributes."""

from .standard import Object, match
from .standard import String, Integer, Float, Boolean
from .containers import Dictionary, List
from .extended import NoneString, NoneInteger, NoneFloat, NoneBoolean
from .extended import Markdown, AttributeDictionary, SortedList



#!/usr/bin/env python
# pylint:disable=R0201

"""Unit tests for the `standard` module."""


import pytest

from yorm.converters import Object, String, Integer, Float, Boolean


class TestObject:

    """Unit tests for the `Object` converter."""

    obj = None

    common = [
        (obj, obj),
        (None, None),
        (1, 1),
        (4.2, 4.2),
        (['a', 'b', 'c'], ['a', 'b', 'c']),
    ]

    data_value = [
    ] + common

    value_data = [
    ] + common

    @pytest.mark.parametrize("data,value", data_value)
    def test_to_value(self, data, value):
        """Verify input data is converted to values."""
        assert value == Object.to_value(data)

    @pytest.mark.parametrize("value,data", value_data)
    def test_to_data(self, value, data):
        """Verify values are converted to output data."""
        assert data == Object.to_data(value)


class TestString:

    """Unit tests for the `String` converter."""

    obj = "Hello, world!"

    common = [
        (obj, obj),
        (None, ""),
        (1, "1"),
        (4.2, "4.2"),
        (['a', 'b', 'c'], "a, b, c"),
    ]

    data_value = [
    ] + common

    value_data = [
    ] + common

    @pytest.mark.parametrize("data,value", data_value)
    def test_to_value(self, data, value):
        """Verify input data is converted to values."""
        assert value == String.to_value(data)

    @pytest.mark.parametrize("value,data", value_data)
    def test_to_data(self, value, data):
        """Verify values are converted to output data."""
        assert data == String.to_data(value)


class TestInteger:

    """Unit tests for the `Integer` converter."""

    obj = 42

    common = [
        (obj, obj),
        (None, 0),
        ("1", 1),
        ("1.1", 1),
    ]

    data_value = [
    ] + common

    value_data = [
    ] + common

    @pytest.mark.parametrize("data,value", data_value)
    def test_to_value(self, data, value):
        """Verify input data is converted to values."""
        assert value == Integer.to_value(data)

    def test_to_value_error(self):
        """Verify an exception is raised for unconvertible values."""
        with pytest.raises(ValueError):
            Integer.to_value("abc")

    @pytest.mark.parametrize("value,data", value_data)
    def test_to_data(self, value, data):
        """Verify values are converted to output data."""
        assert data == Integer.to_data(value)


class TestFloat:

    """Unit tests for the `Float` converter."""

    obj = 4.2

    common = [
        (obj, obj),
        (None, 0.0),
        ("1.0", 1.0),
        ("1.1", 1.1),
    ]

    data_value = [
    ] + common

    value_data = [
    ] + common

    @pytest.mark.parametrize("data,value", data_value)
    def test_to_value(self, data, value):
        """Verify input data is converted to values."""
        assert value == Float.to_value(data)

    def test_to_value_error(self):
        """Verify an exception is raised for unconvertible values."""
        with pytest.raises(ValueError):
            Integer.to_value("abc")

    @pytest.mark.parametrize("value,data", value_data)
    def test_to_data(self, value, data):
        """Verify values are converted to output data."""
        assert data == Float.to_data(value)


class TestBoolean:

    """Unit tests for the `Boolean` converter."""

    obj = True

    common = [
        (obj, obj),
        (None, False),
        (0, False),
        (1, True),
        ("", False),
        ("True", True),
        ("False", False),
        ("true", True),
        ("false", False),
        ("T", True),
        ("F", False),
        ("yes", True),
        ("no", False),
        ("Y", True),
        ("N", False),
        ("enabled", True),
        ("disabled", False),
        ("on", True),
        ("off", False),
        ("Hello, world!", True)
    ]

    data_value = [
    ] + common

    value_data = [
    ] + common

    @pytest.mark.parametrize("data,value", data_value)
    def test_to_value(self, data, value):
        """Verify input data is converted to values."""
        assert value == Boolean.to_value(data)

    @pytest.mark.parametrize("value,data", value_data)
    def test_to_data(self, value, data):
        """Verify values are converted to output data."""
        assert data == Boolean.to_data(value)


if __name__ == '__main__':
    pytest.main()



"""Sample converters and mapped classes for tests."""

from unittest.mock import Mock

from yorm.base.converter import Converter
from yorm.base.mappable import Mappable
from yorm.utilities import sync, attr
from yorm.converters import Dictionary, List
from yorm.converters import String, Integer, Float, Boolean
from yorm.converters import AttributeDictionary, SortedList


# sample converters ###########################################################


class Level(String):

    """Sample custom attribute."""

    @classmethod
    def to_data(cls, obj):
        value = cls.to_value(obj)
        count = value.split('.')
        if count == 0:
            return int(value)
        elif count == 1:
            return float(value)
        else:
            return value


# mock containers #############################################################


class MockConverter(Converter):

    """Sample converter class."""

    @classmethod
    def create_default(cls):
        return None

    @classmethod
    def to_value(cls, _):
        return None

    @classmethod
    def to_data(cls, _):
        return None


class MockConverter0(MockConverter):

    """Sample converter class."""


class MockConverter1(MockConverter):

    """Sample converter class."""


class MockConverter2(MockConverter):

    """Sample converter class."""


class MockConverter3(MockConverter):

    """Sample converter class."""


class MockConverter4(MockConverter):

    """Sample converter class."""

    pass


# sample containers ###########################################################


class EmptyDictionary(Dictionary):

    """Sample dictionary container."""


@attr(all=Integer)
class IntegerList(List):

    """Sample list container."""


@attr(status=Boolean, label=String)
class StatusDictionary(Dictionary):

    """Sample dictionary container."""


@attr(all=StatusDictionary)
class StatusDictionaryList(List):

    """Sample list container."""


@attr(abc=Integer)
class SampleDictionary(Dictionary):

    """Sample dictionary container."""


@attr(var1=Integer, var2=String)
class SampleDictionaryWithInitialization(Dictionary):

    """Sample dictionary container with initialization."""

    def __init__(self, var1, var2, var3):
        super().__init__()
        self.var1 = var1
        self.var2 = var2
        self.var3 = var3


@attr(all=String)
class StringList(List):

    """Sample list container."""


class UnknownList(List):

    """Sample list container."""

    pass


# sample extended containers ##################################################


@attr(var1=Integer, var2=String)
class SampleAttributeDictionary(AttributeDictionary):

    """Sample dictionary container with initialization."""

    def __init__(self, var1, var2, var3):
        super().__init__()
        self.var1 = var1
        self.var2 = var2
        self.var3 = var3


@attr(all=Float)
class SampleSortedList(SortedList):

    """Sample sorted list container."""


class UnknownSortedList(SortedList):

    """Sample list container."""

    pass


# mock mapped classes #########################################################


class MockMappable(Mappable):

    """Sample mappable class."""

    yorm_mapper = Mock()
    yorm_mapper.attrs = {}


# sample mapped classes #######################################################


class SampleStandard:

    """Sample class using standard attribute types."""

    def __init__(self):
        # https://docs.python.org/3.4/library/json.html#json.JSONDecoder
        self.object = {}
        self.array = []
        self.string = ""
        self.number_int = 0
        self.number_real = 0.0
        self.true = True
        self.false = False
        self.null = None

    def __repr__(self):
        return "<standard {}>".format(id(self))


class SampleNested:

    """Sample class using nested attribute types."""

    def __init__(self):
        self.count = 0
        self.results = []

    def __repr__(self):
        return "<nested {}>".format(id(self))


@attr(object=EmptyDictionary, array=IntegerList, string=String)
@attr(number_int=Integer, number_real=Float)
@attr(true=Boolean, false=Boolean)
@sync("path/to/{self.category}/{self.name}.yml")
class SampleStandardDecorated:

    """Sample class using standard attribute types."""

    def __init__(self, name, category='default'):
        self.name = name
        self.category = category
        # https://docs.python.org/3.4/library/json.html#json.JSONDecoder
        self.object = {}
        self.array = []
        self.string = ""
        self.number_int = 0
        self.number_real = 0.0
        self.true = True
        self.false = False
        self.null = None

    def __repr__(self):
        return "<decorated {}>".format(id(self))


@attr(string=String)
@sync("sample.yml", auto=False)
class SampleDecoratedAutoOff:

    """Sample class with automatic storage turned off."""

    def __init__(self):
        self.string = ""

    def __repr__(self):
        return "<auto off {}>".format(id(self))


@attr(string=String, number_real=Float)
class SampleDecoratedNoPath:

    """Sample class with a manually mapped path."""

    def __init__(self):
        self.string = ""
        self.number_real = 0.0

    def __repr__(self):
        return "<no path {}>".format(id(self))


@sync("sample.yml")
class SampleEmptyDecorated:

    """Sample class using standard attribute types."""

    def __repr__(self):
        return "<empty {}>".format(id(self))


class SampleExtended:

    """Sample class using extended attribute types."""

    def __init__(self):
        self.text = ""

    def __repr__(self):
        return "<extended {}>".format(id(self))


@sync("path/to/directory/{UUID}.yml", attrs={'level': Level})
class SampleCustomDecorated:

    """Sample class using custom attribute types."""

    def __init__(self, name):
        self.name = name
        self.level = '1.0'

    def __repr__(self):
        return "<custom {}>".format(id(self))



#!/usr/bin/env python
# pylint:disable=R0201,C0111,R0901

"""Unit tests for the `container` module."""

from unittest.mock import patch

import pytest

import yorm
from yorm import common
from yorm.utilities import attr
from yorm.converters import Dictionary, List
from yorm.converters import String, Integer

from . import strip


@attr(abc=Integer)
class SampleDictionary(Dictionary):

    """Sample dictionary container."""


@attr(var1=Integer, var2=String)
class SampleDictionaryWithInitialization(Dictionary):

    """Sample dictionary container with initialization."""

    def __init__(self, var1, var2, var3):
        super().__init__()
        self.var1 = var1
        self.var2 = var2
        self.var3 = var3


@attr(all=String)
class StringList(List):

    """Sample list container."""


class UnknownList(List):

    """Sample list container."""


class TestDictionary:

    """Unit tests for the `Dictionary` container."""

    obj = {'abc': 123}

    class SampleClass:

        def __init__(self):
            self.abc = 42

    data_value = [
        (obj, obj),
        (None, {'abc': 0}),
        ("key=value", {'key': "value", 'abc': 0}),
        ("key=", {'key': "", 'abc': 0}),
        ("key", {'key': None, 'abc': 0}),
    ]

    value_data = [
        (obj, obj),
        (SampleClass(), {'abc': 42}),
        ([], {'abc': 0}),
    ]

    def setup_method(self, _):
        """Reset the class' mapped attributes before each test."""
        common.ATTRS[SampleDictionary] = {'abc': Integer}

    @pytest.mark.parametrize("data,value", data_value)
    def test_to_value(self, data, value):
        """Verify input data is converted to values."""
        assert value == SampleDictionary.to_value(data)

    @pytest.mark.parametrize("value,data", value_data)
    def test_to_data(self, value, data):
        """Verify values are converted to output data."""
        assert data == SampleDictionary.to_data(value)

    def test_not_implemented(self):
        """Verify `Dictionary` cannot be used directly."""
        with pytest.raises(NotImplementedError):
            Dictionary()

    def test_dict_as_object(self):
        """Verify a `Dictionary` can be used as an attribute."""
        dictionary = SampleDictionaryWithInitialization(1, 2, 3.0)
        value = {'var1': 1, 'var2': '2'}
        value2 = dictionary.to_value(dictionary)
        assert value == value2
        # keys are not accessible as attributes
        assert not hasattr(value2, 'var1')
        assert not hasattr(value2, 'var2')
        assert not hasattr(value2, 'var3')


class TestList:

    """Unit tests for the `List` container."""

    obj = ["a", "b", "c"]

    data_value = [
        (obj, obj),
        (None, []),
        ("a b c", ["a", "b", "c"]),
        ("a,b,c", ["a", "b", "c"]),
        ("abc", ["abc"]),
        ("a\nb\nc", ["a", "b", "c"]),
        (4.2, ['4.2']),
    ]

    value_data = [
        (obj, obj),
    ]

    @pytest.mark.parametrize("data,value", data_value)
    def test_to_value(self, data, value):
        """Verify input data is converted to values."""
        assert value == StringList.to_value(data)

    @pytest.mark.parametrize("value,data", value_data)
    def test_to_data(self, value, data):
        """Verify values are converted to output data."""
        assert data == StringList.to_data(value)

    def test_item_type(self):
        """Verify list item type can be determined."""
        assert String == StringList.item_type

    def test_item_type_none(self):
        """Verify list item type defaults to None."""
        assert None is UnknownList.item_type

    def test_not_implemented(self):
        """Verify `List` cannot be used directly."""
        with pytest.raises(NotImplementedError):
            List()
        with pytest.raises(NotImplementedError):
            UnknownList()


class TestExtensions:

    """Unit tests for extensions to the container classes."""

    class FindMixin:

        def find(self, value):
            for value2 in self:
                if value.lower() == value2.lower():
                    return value2
            return None

    @yorm.attr(a=yorm.converters.String)
    class MyDictionary(Dictionary, FindMixin):
        pass

    @yorm.attr(all=yorm.converters.String)
    class MyList(List, FindMixin):
        pass

    def test_converted_dict_keeps_type(self):
        my_dict = self.MyDictionary()
        my_dict['a'] = 1
        my_dict2 = self.MyDictionary.to_value(my_dict)
        assert 'a' == my_dict2.find('A')
        assert None is my_dict2.find('B')

    def test_converted_list_keeps_type(self):
        my_list = self.MyList()
        my_list.append('a')
        my_list2 = self.MyList.to_value(my_list)
        assert 'a' == my_list2.find('A')
        assert None is my_list2.find('B')


@patch('yorm.settings.fake', True)
class TestReservedNames:

    class MyObject:

        def __init__(self, items=None):
            self.items = items or []

    def test_list_named_items(self):
        obj = self.MyObject()
        yorm.sync_object(obj, "fake/path", {'items': StringList})

        obj.items.append('foo')
        assert strip("""
        items:
        - foo
        """) == obj.yorm_mapper.text

        obj.yorm_mapper.text = strip("""
        items:
        - bar
        """)
        assert ['bar'] == obj.items


if __name__ == '__main__':
    pytest.main()



#!/usr/bin/env python
# pylint: disable=R0201

"""Unit tests for the `mapper` module."""

import os
import pytest
from unittest.mock import patch, Mock

from yorm import mapper
from yorm.converters import Integer


@patch('yorm.settings.fake', True)
class TestHelperFake:

    """Unit tests for fake mappings using the `Helper` class."""

    def test_create(self):
        """Verify fake files can be created."""
        mapped = mapper.Helper("fake/path/to/file")
        mapped.create(Mock())

        assert not os.path.exists(mapped.path)

    def test_delete(self):
        """Verify fake files can be deleted."""
        mapped = mapper.Helper("fake/path/to/file")
        mapped.create(None)
        mapped.delete()

        assert not os.path.exists(mapped.path)

    def test_modified(self):
        """Verify fake files can be modified."""
        mapped = mapper.Helper("fake/path/to/file")
        assert mapped.modified

        mapped.create(None)
        assert mapped.modified

        mapped.modified = False
        assert not mapped.modified

        mapped.modified = True
        assert mapped.modified


class TestHelperReal:

    """Unit tests for real mappings using the `Helper` class."""

    def test_create(self, tmpdir):
        """Verify files can be created."""
        tmpdir.chdir()
        mapped = mapper.Helper("real/path/to/file")
        mapped.create(None)

        assert os.path.isfile(mapped.path)

    def test_create_twice(self, tmpdir):
        """Verify the second creation is ignored."""
        tmpdir.chdir()
        mapped = mapper.Helper("real/path/to/file")
        mapped.create(None)
        mapped.create(None)

        assert os.path.isfile(mapped.path)

    def test_delete(self, tmpdir):
        """Verify files can be deleted."""
        tmpdir.chdir()
        mapped = mapper.Helper("real/path/to/file")
        mapped.create(None)
        mapped.delete()

        assert not os.path.exists(mapped.path)

    def test_delete_twice(self, tmpdir):
        """Verify the second deletion is ignored."""
        tmpdir.chdir()
        mapped = mapper.Helper("real/path/to/file")
        mapped.delete()

        assert not os.path.exists(mapped.path)

    def test_modified(self, tmpdir):
        """Verify files track modifications."""
        tmpdir.chdir()
        mapped = mapper.Helper("real/path/to/file")
        assert mapped.modified

        mapped.create(None)
        assert mapped.modified

        mapped.modified = False
        assert not mapped.modified

        mapped.modified = True
        assert mapped.modified

    def test_modified_deleted(self):
        """Verify a deleted file is always modified."""
        mapped = mapper.Helper("fake/path/to/file")

        assert mapped.modified


class TestMapper:

    """Unit tests for the `Mapper` class."""

    def test_auto_off(self, tmpdir):
        """Verify storage is delayed with auto off."""
        tmpdir.chdir()
        attrs = {'number': Integer}
        mapped = mapper.Mapper(None, "real/path/to/file", attrs, auto=False)
        assert False is mapped.auto

        mapped.create()
        assert "" == mapped.text
        assert False is mapped.auto

        mapped.store()
        assert "" == mapped.text
        assert False is mapped.auto

        mapped.store(force=True)
        assert "number: 0\n" == mapped.text
        assert False is mapped.auto


if __name__ == '__main__':
    pytest.main()



"""Functions and decorators."""

import uuid

from . import common
from .base.mappable import MAPPER, Mappable
from .mapper import Mapper

log = common.logger(__name__)

UUID = 'UUID'


def sync(*args, **kwargs):
    """Convenience function to forward calls based on arguments.

    This function will call either:

    * `sync_object` - when given an unmapped object
    * `sync_instances` - when used as the class decorator

    Consult the signature of each call for more information.

    """
    if 'path_format' in kwargs or args and isinstance(args[0], str):
        return sync_instances(*args, **kwargs)
    else:
        return sync_object(*args, **kwargs)


def sync_object(instance, path, attrs=None, auto=True):
    """Enable YAML mapping on an object.

    :param instance: object to patch with YAML mapping behavior
    :param path: file path for dump/load
    :param attrs: dictionary of attribute names mapped to converter classes
    :param auto: automatically store attribute to file

    """
    log.info("mapping object...")
    _check_base(instance, mappable=False)

    attrs = attrs or common.ATTRS[instance.__class__]

    class Mapped(Mappable, instance.__class__):

        """Original class with `Mappable` as the base."""

    mapper = Mapper(instance, path, attrs, auto=auto)

    if not mapper.exists:
        mapper.create()
        mapper.store(force=True)
    mapper.fetch(force=True)

    setattr(instance, MAPPER, mapper)
    instance.__class__ = Mapped
    log.info("mapped %r to '%s'", instance, path)

    return instance


def sync_instances(path_format, format_spec=None, attrs=None, auto=True):
    """Class decorator to enable YAML mapping after instantiation.

    :param path_format: formatting string to create file paths for dump/load
    :param format_spec: dictionary to use for string formatting
    :param attrs: dictionary of attribute names mapped to converter classes
    :param auto: automatically store attribute to file

    """
    format_spec = format_spec or {}
    attrs = attrs or {}

    def decorator(cls):
        """Class decorator."""

        class Mapped(Mappable, cls):

            """Original class with `Mappable` as the base."""

            def __init__(self, *_args, **_kwargs):
                setattr(self, MAPPER, None)
                super().__init__(*_args, **_kwargs)

                log.info("mapping instance of %r to '%s'...", cls, path_format)
                format_values = {}
                for key, value in format_spec.items():
                    format_values[key] = getattr(self, value)
                if '{' + UUID + '}' in path_format:
                    format_values[UUID] = uuid.uuid4().hex
                format_values['self'] = self

                path = path_format.format(**format_values)
                attrs.update(common.ATTRS[self.__class__])
                attrs.update(common.ATTRS[cls])
                mapper = Mapper(self, path, attrs, auto=auto)

                if not mapper.exists:
                    mapper.create()
                    mapper.store(force=True)
                mapper.fetch(force=True)

                setattr(self, MAPPER, mapper)
                log.info("mapped %r to '%s'", self, path)

        return Mapped

    return decorator


def attr(**kwargs):
    """Class decorator to map attributes to converters.

    :param kwargs: keyword arguments mapping attribute name to converter class

    """
    def decorator(cls):
        """Class decorator."""
        for name, converter in kwargs.items():
            common.ATTRS[cls][name] = converter

        return cls

    return decorator


def update(instance, fetch=True, force=True, store=True):
    """Synchronize changes between a mapped object and its file.

    :param instance: object with patched YAML mapping behavior
    :param fetch: update the object with changes from its file
    :param force: even if the file appears unchanged
    :param store: update the file with changes from the object

    """
    _check_base(instance, mappable=True)

    if fetch:
        update_object(instance, force=False)
    if store:
        update_file(instance)
    if fetch:
        update_object(instance, force=force)


def update_object(instance, force=True):
    """Synchronize changes into a mapped object from its file.

    :param instance: object with patched YAML mapping behavior
    :param force: update the object even if the file appears unchanged

    """
    log.info("manually updating %r from file...", instance)
    _check_base(instance, mappable=True)

    mapper = getattr(instance, MAPPER)
    mapper.fetch(force=force)


def update_file(instance, force=True):
    """Synchronize changes into a mapped object's file.

    :param instance: object with patched YAML mapping behavior
    :param force: update the file even if automatic sync is off

    """
    log.info("manually saving %r to file...", instance)
    _check_base(instance, mappable=True)

    mapper = getattr(instance, MAPPER)
    mapper.store(force=force)


def _check_base(obj, mappable=True):
    """Confirm an object's base class is `Mappable` as required."""
    if mappable and not isinstance(obj, Mappable):
        raise common.UseageError("{} is not mapped".format(repr(obj)))
    if not mappable and isinstance(obj, Mappable):
        raise common.UseageError("{} is already mapped".format(repr(obj)))
